/**
 * Copyright(c) 2020-present, Odysseas Georgoudis & quill contributors.
 * Distributed under the MIT License (http://opensource.org/licenses/MIT)
 */

#pragma once

#include "quill/detail/misc/Attributes.h"
#include "quill/detail/misc/Common.h"

#include "quill/Fmt.h"
#include "quill/LogLevel.h"
#include "quill/QuillError.h"
#include "quill/clock/TimestampClock.h"
#include "quill/detail/LoggerDetails.h"
#include "quill/detail/Serialize.h"
#include "quill/detail/ThreadContext.h"
#include "quill/detail/ThreadContextCollection.h"
#include "quill/detail/misc/Rdtsc.h"
#include "quill/detail/misc/TypeTraitsCopyable.h"
#include "quill/detail/misc/Utilities.h"
#include <atomic>
#include <cstdint>
#include <thread>
#include <vector>

namespace quill
{

namespace detail
{
class LoggerCollection;
class LogManager;
} // namespace detail

/**
 * Thread safe logger.
 * Logger must be obtained from LoggerCollection get_logger(), therefore constructors are private
 */
class alignas(detail::CACHE_LINE_ALIGNED) Logger
{
public:
  /**
   * Deleted
   */
  Logger(Logger const&) = delete;
  Logger& operator=(Logger const&) = delete;

  /**
   * We align the logger object to it's own cache line. It shouldn't make much difference as the
   * logger object size is exactly 1 cache line
   */
  void* operator new(size_t i) { return detail::alloc_aligned(i, detail::CACHE_LINE_ALIGNED); }
  void operator delete(void* p) { detail::free_aligned(p); }

  /**
   * @return The log level of the logger
   */
  QUILL_NODISCARD LogLevel log_level() const noexcept
  {
    return _log_level.load(std::memory_order_relaxed);
  }

  /**
   * Set the log level of the logger
   * @param log_level The new log level
   */
  void set_log_level(LogLevel log_level)
  {
    if (QUILL_UNLIKELY(log_level == LogLevel::Backtrace))
    {
      QUILL_THROW(QuillError{"LogLevel::Backtrace is only used internally. Please don't use it."});
    }

    _log_level.store(log_level, std::memory_order_relaxed);
  }

  /**
   * Checks if the given log_statement_level can be logged by this logger
   * @tparam log_statement_level The log level of the log statement to be logged
   * @return bool if a message can be logged based on the current log level
   */
  template <LogLevel log_statement_level>
  QUILL_NODISCARD_ALWAYS_INLINE_HOT bool should_log() const noexcept
  {
    if constexpr (static_cast<LogLevel>(QUILL_ACTIVE_LOG_LEVEL) > log_statement_level)
    {
      return false;
    }

    return log_statement_level >= log_level();
  }

  /**
   * Checks if the given log_statement_level can be logged by this logger
   * @param log_statement_level The log level of the log statement to be logged
   * @return bool if a message can be logged based on the current log level
   */
  QUILL_NODISCARD_ALWAYS_INLINE_HOT bool should_log(LogLevel log_statement_level) const noexcept
  {
    if constexpr (QUILL_ACTIVE_LOG_LEVEL > 0)
    {
      if (static_cast<LogLevel>(QUILL_ACTIVE_LOG_LEVEL) > log_statement_level)
      {
        return false;
      }
    }

    return log_statement_level >= log_level();
  }

  /**
   * Push a log message to the spsc queue to be logged by the backend thread.
   * One spsc queue per caller thread. This function is enabled only when all arguments are
   * fundamental types.
   * This is the fastest way possible to log
   * @note This function is thread-safe.
   * @param dynamic_log_level dynamic log level
   * @param fmt_args arguments
   */
  template <bool IsPrintfFormat, typename... Args>
  QUILL_ALWAYS_INLINE_HOT void log(LogLevel dynamic_log_level, MacroMetadata const* macro_metadata, Args&&... fmt_args)
  {
    assert(!_is_invalidated.load(std::memory_order_acquire) && "Invalidated loggers can not log");

#if QUILL_FMT_VERSION >= 90000
    static_assert(!detail::has_fmt_stream_view_v<Args...>,
                  "fmtquill::streamed(...) is not supported. In order to make a type formattable "
                  "via std::ostream you should provide a formatter specialization inherited "
                  "from ostream_formatter. "
                  "`template <> struct fmtquill::formatter<T> : ostream_formatter {};");
#endif

#if !defined(QUILL_MODE_UNSAFE)
    {
      // not allowing unsafe copies
      static_assert(detail::are_copyable_v<Args...>,
                    "Trying to copy an unsafe to copy type. Tag or specialize the type as "
                    "`copy_loggable` or explicitly format the type to string on the caller thread"
                    "prior to logging. See "
                    "https://github.com/odygrd/quill/wiki/8.-User-Defined-Types for more info.");
    }
#endif

    // Store the timestamp of the log statement at the start of the call. This gives more accurate
    // timestamp especially if the queue is full
    uint64_t const timestamp = (_logger_details.timestamp_clock_type() == TimestampClockType::Tsc)
      ? detail::rdtsc()
      : (_logger_details.timestamp_clock_type() == TimestampClockType::System)
      ? static_cast<uint64_t>(std::chrono::duration_cast<std::chrono::nanoseconds>(
                                std::chrono::system_clock::now().time_since_epoch())
                                .count())
      : _custom_timestamp_clock->now();

    detail::ThreadContext* const thread_context =
      _thread_context_collection.local_thread_context<QUILL_QUEUE_TYPE>();

    // For windows also take wide strings into consideration.
#if defined(_WIN32)
    constexpr uint32_t c_string_count = detail::count_c_style_strings<Args...>() +
      detail::count_c_style_wide_strings<Args...>() + detail::count_std_wstring_type<Args...>();
#else
    constexpr uint32_t c_string_count = detail::count_c_style_strings<Args...>();
#endif

    size_t c_string_sizes[(std::max)(c_string_count, static_cast<uint32_t>(1))];

    // Need to reserve additional space as we will be aligning the pointer
    size_t total_size = alignof(uint64_t) + sizeof(uint64_t) + (sizeof(uintptr_t) * 3) +
      detail::calculate_args_size_and_populate_string_lengths(c_string_sizes, fmt_args...);

    if (dynamic_log_level != LogLevel::None)
    {
      // For the dynamic log level we want to add to the total size to store the dynamic log level
      total_size += sizeof(LogLevel);
    }

    // request this size from the queue
    std::byte* write_buffer =
      thread_context->spsc_queue<QUILL_QUEUE_TYPE>().prepare_write(static_cast<uint32_t>(total_size));

    if constexpr (QUILL_QUEUE_TYPE == detail::QueueType::UnboundedNoMaxLimit)
    {
      assert(write_buffer && "UnboundedNonBlocking will always allocate and get new space");
    }
    else if constexpr ((QUILL_QUEUE_TYPE == detail::QueueType::BoundedNonBlocking) ||
                       (QUILL_QUEUE_TYPE == detail::QueueType::UnboundedDropping))
    {
      if (QUILL_UNLIKELY(write_buffer == nullptr))
      {
        // not enough space to push to queue message is dropped
        thread_context->increment_message_failure_counter();
        return;
      }
    }
    else if constexpr ((QUILL_QUEUE_TYPE == detail::QueueType::BoundedBlocking) ||
                       (QUILL_QUEUE_TYPE == detail::QueueType::UnboundedBlocking))
    {
      if (QUILL_UNLIKELY(write_buffer == nullptr))
      {
        thread_context->increment_message_failure_counter();

        do
        {
          if constexpr (QUILL_BLOCKING_QUEUE_RETRY_INTERVAL_NS > 0)
          {
            std::this_thread::sleep_for(std::chrono::nanoseconds{QUILL_BLOCKING_QUEUE_RETRY_INTERVAL_NS});
          }

          // not enough space to push to queue, keep trying
          write_buffer =
            thread_context->spsc_queue<QUILL_QUEUE_TYPE>().prepare_write(static_cast<uint32_t>(total_size));
        } while (write_buffer == nullptr);
      }
    }

    // we have enough space in this buffer, and we will write to the buffer

    // Then write the pointer to the LogDataNode. The LogDataNode has all details on how to
    // deserialize the object. We will just serialize the arguments in our queue, but we need
    // to look up their types to deserialize them

    // Note: The metadata variable here is created during program init time,
    std::byte const* const write_begin = write_buffer;
    write_buffer = detail::align_pointer<alignof(uint64_t), std::byte>(write_buffer);

    std::memcpy(write_buffer, &timestamp, sizeof(timestamp));
    write_buffer += sizeof(timestamp);

    std::memcpy(write_buffer, &macro_metadata, sizeof(uintptr_t));
    write_buffer += sizeof(uintptr_t);

    detail::LoggerDetails const* logger_details = &_logger_details;
    std::memcpy(write_buffer, &logger_details, sizeof(uintptr_t));
    write_buffer += sizeof(uintptr_t);

    if constexpr (IsPrintfFormat)
    {
      detail::PrintfFormatToFn ftf = detail::printf_format_to<Args...>;
      std::memcpy(write_buffer, &ftf, sizeof(uintptr_t));
      write_buffer += sizeof(uintptr_t);
    }
    else
    {
      detail::FormatToFn ftf = detail::format_to<Args...>;
      std::memcpy(write_buffer, &ftf, sizeof(uintptr_t));
      write_buffer += sizeof(uintptr_t);
    }

    // encode remaining arguments
    detail::encode(write_buffer, c_string_sizes, fmt_args...);

    if (dynamic_log_level != LogLevel::None)
    {
      // write the dynamic log level
      std::memcpy(write_buffer, &dynamic_log_level, sizeof(LogLevel));
      write_buffer += sizeof(LogLevel);
    }

    assert(total_size >= (static_cast<uint32_t>(write_buffer - write_begin)) &&
           "The committed write bytes can not be greater than the requested bytes");
    assert((write_buffer >= write_begin) &&
           "write_buffer should be greater or equal to write_begin");
    thread_context->spsc_queue<QUILL_QUEUE_TYPE>().finish_write(static_cast<uint32_t>(write_buffer - write_begin));
    thread_context->spsc_queue<QUILL_QUEUE_TYPE>().commit_write();
  }

  /**
   * Init a backtrace for this logger.
   * Stores messages logged with LOG_BACKTRACE in a ring buffer messages and displays them later on demand.
   * @param capacity The max number of messages to store in the backtrace
   * @param backtrace_flush_level If this loggers logs any message higher or equal to this severity level the backtrace will also get flushed.
   * Default level is None meaning the user has to call flush_backtrace explicitly
   */
  void init_backtrace(uint32_t capacity, LogLevel backtrace_flush_level = LogLevel::None)
  {
    assert(!_is_invalidated.load(std::memory_order_acquire) &&
           "Invalidated loggers can not be used");

    // we do not care about the other fields, except quill::MacroMetadata::Event::InitBacktrace
    static constexpr MacroMetadata macro_metadata{
      "",    "",   "{}", nullptr, LogLevel::Critical, MacroMetadata::Event::InitBacktrace,
          false, false};

    // we pass this message to the queue and also pass capacity as arg
    this->template log<false>(LogLevel::None, &macro_metadata, capacity);

    // Also store the desired flush log level
    _logger_details.set_backtrace_flush_level(backtrace_flush_level);
  }

  /**
   * Dump any stored backtrace messages
   */
  void flush_backtrace()
  {
    assert(!_is_invalidated.load(std::memory_order_acquire) &&
           "Invalidated loggers can not be used");

    // we do not care about the other fields, except quill::MacroMetadata::Event::Flush
    static constexpr MacroMetadata macro_metadata{
      "",    "",   "", nullptr, LogLevel::Critical, MacroMetadata::Event::FlushBacktrace,
          false, false};

    // we pass this message to the queue
    this->template log<false>(LogLevel::None, &macro_metadata);
  }

  /**
   * @return The name of the logger
   */
  QUILL_NODISCARD std::string const& name() const noexcept { return _logger_details.name(); }

private:
  friend class detail::LoggerCollection;
  friend class detail::LogManager;

  /**
   * Constructs new logger object
   * @param name the name of the logger
   * @param handler handlers for this logger
   * @param timestamp_clock_type timestamp clock
   * @param custom_timestamp_clock custom timestamp clock
   * @param thread_context_collection thread context collection reference
   */
  Logger(std::string const& name, std::shared_ptr<Handler> handler, TimestampClockType timestamp_clock_type,
         TimestampClock* custom_timestamp_clock, detail::ThreadContextCollection& thread_context_collection)
    : _logger_details(name, std::move(handler), timestamp_clock_type),
      _custom_timestamp_clock(custom_timestamp_clock),
      _thread_context_collection(thread_context_collection)
  {
    if ((timestamp_clock_type == TimestampClockType::Custom) && !custom_timestamp_clock)
    {
      QUILL_THROW(
        QuillError{"A valid TimestampClock* needs to be provided when TimestampClockType is set to "
                   "Custom. Call 'quill::set_custom_timestamp_clock(...)'"});
    }
  }

  /**
   * Constructs a new logger object with multiple handlers
   */
  Logger(std::string const& name, std::vector<std::shared_ptr<Handler>> const& handlers,
         TimestampClockType timestamp_clock_type, TimestampClock* custom_timestamp_clock,
         detail::ThreadContextCollection& thread_context_collection)
    : _logger_details(name, handlers, timestamp_clock_type),
      _custom_timestamp_clock(custom_timestamp_clock),
      _thread_context_collection(thread_context_collection)
  {
    if ((timestamp_clock_type == TimestampClockType::Custom) && !custom_timestamp_clock)
    {
      QUILL_THROW(
        QuillError{"A valid TimestampClock* needs to be provided when TimestampClockType is set to "
                   "Custom. Call 'quill::set_custom_timestamp_clock(...)'"});
    }
  }

  void invalidate() { _is_invalidated.store(true, std::memory_order_release); }

  QUILL_NODISCARD bool is_invalidated() const noexcept
  {
    return _is_invalidated.load(std::memory_order_acquire);
  }

private:
  detail::LoggerDetails _logger_details;
  TimestampClock* _custom_timestamp_clock{nullptr}; /* A non owned pointer to a custom timestamp clock, valid only when provided */
  detail::ThreadContextCollection& _thread_context_collection;
  std::atomic<LogLevel> _log_level{LogLevel::Info};
  std::atomic<bool> _is_invalidated{false};
};

} // namespace quill
